% conjgrad - conjugate gradient algorithm.
%
% This routine iteratively solves the linear system:
%
%     A.x = b
%
% Arguments:
%    A   - The left-hand-side operator, it is the name or a function handle
%          of a user defined function which takes a single argument and
%          returns operator `A` applied to this argument.
%    b   - The right-hand-side vector of the normal equations.
%    x0  - The initial guess for the solution `x`.
%    tol - The tolerance(s) for convergence, can be one or two values: `atol`
%           (as if `rtol` = 0) or [`atol`, `rtol`] where `atol` and `rtol` are
%           the absolute and relative tolerances.  Convergence occurs when the
%           Euclidean norm of the residuals is less or equal the largest of
%           `atol` and `rtol` times the Eucliden norm of the initial
%           residuals.
%    maxiter - The maximum number of iterations.
%
function x = conjgrad(A, b, x0, tol, maxiter)

    % Resolve function handle.
    if ischar(A)
        A = str2func(A);
    end

    % Initialization.
    x = x0;
    r = b - A(x);
    rho = conjgrad_inner(r, r);
    if length(tol) == 1
        epsilon = tol(1);
    elseif length(tol) == 2
        epsilon = max(tol(1), tol(2)*sqrt(rho));
    end

    % Conjugate gradient iterations.
    k = 0;
    rho0 = 0.0;
    while 1
        k =  k + 1;
        if sqrt(rho) <= epsilon
            % Normal convergence.
            break
        elseif k > maxiter
            fprintf('WARNING - too many (%d) conjugate gradient iterations\n', ...
                   maxiter);
            break
        end
        if k == 1
            % First iteration.
            p = r;
        else
            p = r + (rho/rho0)*p;
        end
        q = A(p);
        gamma = conjgrad_inner(p, q);
        if gamma <= 0.0
            error('left-hand-side operator A is not positive definite');
            break
        end
        alpha = rho/gamma;
        x = x + alpha*p;
        r = r - alpha*q;
        rho0 = rho;
        rho = conjgrad_inner(r, r);
    end
end

% conjgrad_inner - Compute the inner product for the conjugate gradient
%                  algorithm.
function s = conjgrad_inner(x, y)
  s = sum(bsxfun(@times, x(:), y(:)));
end
